@@ -9,6 +9,8 @@
  * 2003-11-16 Break out oem data decoder so each platform can handle the
  *	      oem data as it likes.
  *	      Keith Owens <kaos@sgi.com>
+ * 2004-10-04 Handle kernels that clear the bit themselves when there is no data.
+ *	      Keith Owens <kaos@sgi.com>
  *
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  *
@@ -97,11 +99,36 @@ usage (void)
 		, stderr);
 }
 
+/* Ben Woodard of RedHat changed the kernel salinfo code around 2.6.9-rc3 to
+ * clear the cpu state bit if there is no data.  He did not add any indication
+ * to user space of this change, which means that user space must deduce if it
+ * is running on a kernel with or without Ben Woodward's change :(.
+ *
+ * Start off by assuming we are running on a changed kernel, and do not write
+ * 'clear' to the kernel when there is no data.  When running on an old kernel,
+ * user space will then be invoked repeatedly with no data.  Detect this loop
+ * for an old kernel and turn on do_clear.
+ */
+
 static int
-clear_cpu(int fd_data, int cpu, const char *data_filename)
+clear_cpu(int fd_data, int cpu, const char *data_filename, int have_data)
 {
 	char text[400];
 	int l;
+	static int prev_cpu = -1, loop = 0, do_clear = 0;
+
+	if (have_data)
+		loop = 0;
+	if (!do_clear) {
+		if (cpu <= prev_cpu) {
+			++loop;
+			if (loop == 2)
+				do_clear = 1;
+		}
+		prev_cpu = cpu;
+	}
+	if (!have_data && !do_clear)
+		return 0;
 
 	snprintf(text, sizeof(text), "clear %d\n", cpu);
 	l = strlen(text);
@@ -155,7 +182,6 @@ fork_oemdata(void)
 		close(pc[1]);
 		oemdata_fd[0] = pc[0];	/* mca.c read from pgm */
 		oemdata_fd[1] = pp[1];	/* mca.c write to pgm */
-		printf("oemdata_fd[0](%d) oemdata_fd[1](%d)\n", oemdata_fd[0], oemdata_fd[1]);
 		__asm__ __volatile__ ("" ::: "memory");	/* lock free barrier */
 		if (child_died)
 			oemdata_fd[1] = -1;
@@ -227,7 +253,7 @@ talk_to_sal (const char *type, const char *directory)
 			goto out;
 		}
 		if (!(buffer = salinfo_buffer(fd_data, &bufsize))) {
-			if (clear_cpu(fd_data, cpu, data_filename))
+			if (clear_cpu(fd_data, cpu, data_filename, 0))
 				goto out;
 			continue;	/* event but no data is normal at boot */
 		}
@@ -278,12 +304,12 @@ talk_to_sal (const char *type, const char *directory)
 			type,
 			suffix);
 		fclose(stdout);
-		if (!(stdout = fopen(filename, "w"))) {
+		if (!(stdout = fopen(filename, "a"))) {
 			perror(filename);
 			goto out;
 		}
 		fclose(stderr);
-		if (!((stderr = fopen(filename, "w")))) {
+		if (!((stderr = fopen(filename, "a")))) {
 			perror(filename);
 			goto out;
 		}
@@ -293,7 +319,7 @@ talk_to_sal (const char *type, const char *directory)
 		printf("END HARDWARE ERROR STATE from %s on cpu %d\n", type, cpu);
 		free(buffer);
 		fclose(stdout);
-		if (clear_cpu(fd_data, cpu, data_filename))
+		if (clear_cpu(fd_data, cpu, data_filename, 1))
 			goto out;
 
 	}
